import logging
from http.cookies import Morsel
from os import environ, getenv
from typing import Mapping

from jwt import InvalidTokenError, InvalidSignatureError, decode
from jwt import PyJWKClient

logger = logging.getLogger(__name__)


def clean_username(username: str) -> str:
    """
    Remove @{suffix} from username, where suffix is `ALLOWEDFLARE_EMAIL_DOMAIN` or
    `ALLOWEDFLARE_PRIVATE_DOMAIN`. Set `ALLOWEDFLARE_EMAIL_DOMAIN=off` to leave the
    username unmodified.

    Compared to `RemoteUserBackend.clean_username()`, the `self` argument is omitted to make
    user-provided replacement easier.
    """
    suffix = getenv('ALLOWEDFLARE_EMAIL_DOMAIN', getenv('ALLOWEDFLARE_PRIVATE_DOMAIN', 'off'))
    return username.removesuffix(f'@{suffix}')


def authenticate(cookies: Mapping[str, Morsel | str]) -> tuple[str, str, dict]:
    """
    Return a tuple with suffix-trimmed username, failure/success message, and decoded token.

    On failure to authenticate, the username will be the empty string `''` and the token will be the
    empty dictionary `{}`.
    """
    url = getenv('ALLOWEDFLARE_ACCESS_URL', 'off')
    if url == 'off':
        return ('', 'Allowedflare is off', {})

    if 'CF_Authorization' not in cookies:
        return ('', 'Allowedflare could not find CF_Authorization cookie', {})

    morsel_or_string = cookies['CF_Authorization']
    if isinstance(morsel_or_string, Morsel):
        cookie = morsel_or_string.value
    else:
        cookie = morsel_or_string

    token = {'token': 'not parsed'}
    try:
        # TODO make this module-level and ensure cache is being used to smooth over intermittent network outages
        jwks_client = PyJWKClient(f'{url}/cdn-cgi/access/certs', timeout=3)
        signing_key = jwks_client.get_signing_key_from_jwt(cookie)
        token = decode(
            cookie,
            key=signing_key.key,
            audience=environ['ALLOWEDFLARE_AUDIENCE'],
            algorithms=['RS256'],
        )
    except InvalidTokenError as error:
        return ('', f'Allowedflare failed to decode CF_Authorization cookie {cookie} {error}', {})
    except InvalidSignatureError:
        return ('', f'Allowedflare found invalid signature in CF_Authorization cookie {token}', {})
    except Exception as error:
        return ('', f'Allowedflare unexpected {error} {token}', {})

    # It looks like people will have email, service tokens will have common_name
    unprocessed_name = token.get('email', token.get('common_name', ''))
    if not unprocessed_name:
        return (
            '',
            f'Allowedflare found neither common_name nor email in otherwise valid token {token}',
            {},
        )

    name = clean_username(unprocessed_name)
    return (name, f'Allowedflare authenticated {name}', token)



import logging
from typing import Any

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.models import User, Group, Permission
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.http import HttpRequest, HttpResponse

# Can be shortened to django.http.HttpResponseBase in 4.1 and newer
# https://github.com/django/django/pull/15667/commits/7e003fa06a40c72f7f442876e1eb2611115df80c
from django.http.response import HttpResponseBase
from rest_framework.authentication import BaseAuthentication

from allowedflare.allowedflare import authenticate

logger = logging.getLogger(__name__)


def configure_user(user: User, request: HttpRequest, created: bool) -> User:
    """
    Ensure the user can view all models through the admin site.

    To match `RemoteUserBackend.configure_user()`, `request` and `created` arguments are accepted
    but unused. The `self` argument is omitted to make user-provided replacement easier.
    """
    user.is_staff = True
    user.save()

    everyone, _ = Group.objects.get_or_create(name='allowedflare_everyone')
    everyone.permissions.add(*Permission.objects.filter(codename__startswith='view'))
    user.groups.add(everyone)
    return user


def update_or_create_user(username: str, request: HttpRequest, token: dict) -> User:
    if 'email' in token:
        defaults = {'email': token['email']}
    else:
        defaults = {}
    user, created = User.objects.update_or_create(username=username, defaults=defaults)
    getattr(settings, 'ALLOWEDFLARE_CONFIGURE_USER', configure_user)(user, request, created)
    return user


class Backend(ModelBackend):
    def authenticate(
        self,
        request: HttpRequest | None,
        username: str | None = None,
        password: str | None = None,
        **kwargs: Any,
    ) -> User | None:
        if not request:
            return None
        username_from_token, message, token = authenticate(request.COOKIES)
        if not username_from_token:
            logger.info(message)
            return None
        # Require an explicit username so that someone with a valid token can still log in to other
        # accounts using ModelBackend or other authentication backends.
        if username != username_from_token:
            logger.info(
                f'Allowedflare found {username} != {username_from_token} (ignoring {message})'
            )
            return None
        logger.info(message)
        return update_or_create_user(username, request, token)

    def get_user(self, user_id: int) -> User | None:
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None


class LoginView(DjangoLoginView):
    template_name = 'login.html'

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        self.username, message, _ = authenticate(self.request.COOKIES)
        self.extra_context = {**(self.extra_context or {}), 'allowedflare_message': message}
        return super().get(request, *args, **kwargs)

    def get_form(self, form_class: type[AuthenticationForm] | None = None) -> AuthenticationForm:
        form = super().get_form(form_class)
        if getattr(self, 'username', False):
            form.fields['password'].initial = 'placeholder'
            form.fields['password'].widget.render_value = True
            form.fields['username'].initial = self.username
        return form


def login_view_wrapper(request: HttpRequest) -> HttpResponseBase:
    return LoginView.as_view(
        extra_context={'title': 'Log In', REDIRECT_FIELD_NAME: request.get_full_path()}
    )(request)


class Authentication(BaseAuthentication):
    def authenticate(self, request: HttpRequest) -> tuple[User | None, dict | None]:
        username, message, token = authenticate(request.COOKIES)
        logger.info(message)
        if not username:
            return (None, None)
        return (update_or_create_user(username, request, token), token)



from jwt import encode
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.backends import default_backend
from datetime import UTC, datetime

from allowedflare import clean_username
from allowedflare.allowedflare import authenticate


def test_clean_username_unmodified(monkeypatch):
    monkeypatch.delenv('ALLOWEDFLARE_EMAIL_DOMAIN', raising=False)
    monkeypatch.delenv('ALLOWEDFLARE_PRIVATE_DOMAIN', raising=False)
    assert clean_username('user@domain.com') == 'user@domain.com'

    monkeypatch.setenv('ALLOWEDFLARE_EMAIL_DOMAIN', 'off')
    monkeypatch.setenv('ALLOWEDFLARE_PRIVATE_DOMAIN', 'domain.com')
    assert clean_username('user@domain.com') == 'user@domain.com'


def test_clean_username_email_domain_removed(monkeypatch):
    monkeypatch.setenv('ALLOWEDFLARE_EMAIL_DOMAIN', 'domain.com')

    assert clean_username('user@domain.com') == 'user'


def test_clean_username_private_domain_removed(monkeypatch):
    monkeypatch.setenv('ALLOWEDFLARE_PRIVATE_DOMAIN', 'domain.com')
    assert clean_username('user@domain.com') == 'user'


def test_authenticate_off(monkeypatch):
    monkeypatch.setenv('ALLOWEDFLARE_ACCESS_URL', 'off')
    assert authenticate({}) == ('', 'Allowedflare is off', {})


# TODO make these offline


# TODO check Cf-Access-Jwt-Assertion header as fallback, or whatever `cloudflared curl` sets
def test_authenticate_no_cookie(monkeypatch):
    monkeypatch.setenv('ALLOWEDFLARE_ACCESS_URL', 'https://demo.cloudflareaccess.com')
    assert authenticate({}) == ('', 'Allowedflare could not find CF_Authorization cookie', {})


def test_authenticate_invalid_token(monkeypatch):
    monkeypatch.setenv('ALLOWEDFLARE_ACCESS_URL', 'https://demo.cloudflareaccess.com')
    user, message, token = authenticate({'CF_Authorization': 'invalid'})
    assert user == ''
    assert message.startswith('Allowedflare failed to decode CF_Authorization cookie invalid')
    assert token == {}


def test_authenticate_jwk_client_error(monkeypatch):
    private_key = generate_private_key(65537, 512, default_backend())
    monkeypatch.setenv('ALLOWEDFLARE_ACCESS_URL', 'https://demo.cloudflareaccess.com')
    user, message, token = authenticate({'CF_Authorization': encode({}, private_key, 'RS256')})
    assert user == ''
    assert 'Unable to find a signing key that matches' in message
    assert token == {}


def test_authenticate_valid_token(mocker, monkeypatch):
    private_key = generate_private_key(65537, 512, default_backend())
    token = {
        'aud': 'audience',
        'email': 'firstname.lastname@domain.com',
        'exp': int(datetime.now(UTC).timestamp() + 100),
    }
    cookies = {'CF_Authorization': encode(token, private_key, 'RS256')}
    monkeypatch.setenv('ALLOWEDFLARE_ACCESS_URL', 'https://demo.cloudflareaccess.com')
    monkeypatch.setenv('ALLOWEDFLARE_AUDIENCE', 'audience')
    mock_get_signing_key_from_jwt = mocker.patch(
        'allowedflare.allowedflare.PyJWKClient.get_signing_key_from_jwt', autospec=True
    )
    mock_get_signing_key_from_jwt.return_value.key = private_key.public_key()

    assert authenticate(cookies) == (
        'firstname.lastname@domain.com',
        'Allowedflare authenticated firstname.lastname@domain.com',
        token,
    )



from django.test import RequestFactory

from allowedflare import LoginView


def test_allowedflare_login_view(monkeypatch, rf: RequestFactory):
    monkeypatch.setenv('ALLOWEDFLARE_ACCESS_URL', 'off')
    response = LoginView.as_view()(rf.get(''))
    assert response.status_code == 200
    assert 'allowedflare_message' in response.context_data  # type: ignore[attr-defined]



from allowedflare.allowedflare import clean_username

# Django REST Framework authentication class
from allowedflare.django import Authentication

# Django Admin authentication backend
from allowedflare.django import Backend

# Django Admin login view
from allowedflare.django import LoginView

# SQL Explorer login view wrapper
from allowedflare.django import login_view_wrapper

__all__ = (
    Authentication.__name__,
    Backend.__name__,
    LoginView.__name__,
    clean_username.__name__,
    login_view_wrapper.__name__,
)



from jupyterhub.auth import Authenticator  # type: ignore
from tornado.web import RequestHandler

from allowedflare.allowedflare import authenticate


class JupyterHub(Authenticator):
    auto_login = True

    # New configurable trait

    async def get_authenticated_user(self, handler: RequestHandler, data):
        username, message, token = authenticate(handler.request.cookies)
        self.log.info(message)
        if not username:
            return None
        return {
            'name': username,
            'admin': True,  # TODO look at groups or something
        }
